"""Response object manages fixtures and does diff."""

import inspect
import json
import os
import subprocess
import tempfile

from bs4 import BeautifulSoup

from .exceptions import DiffsFound


def diff(first, second):
    """Return the command and diff output between first and second paths."""
    cmd = 'diff -u1 %s %s | sed "1,2 d"' % (first, second)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    out, err = proc.communicate()
    return cmd, out


class Response(object):
    """
    Object to use in tests.

    Consider this example:

    .. code-block:: python

        class TestYourView(TestCase):
            def test_your_page(self):
                result = test.Client().get(your_url)

                # Factory to create a Response for this test
                expected = Response.for_test(self)

                # Generate the fixture if necessary, otherwise GNU diff-it
                expected.assertNoDiff(result)

    This would create a directory in the directory containing TestYourView,
    named ``responsediff_fixtures``, with a sub-directory
    ``TestYourView.test_your_page`` and a file ``content`` with
    response.content in there, and another called ``metadata`` with additional
    information about the response in there such as the status code. The test
    would fail with a DiffsFound error containing the list of created files to
    inform the user that no test has actually been run, and that the fixture
    has just been created.

    User should add the generated fixture to the repository. Then, next time
    this test is run, it will run GNU diff between ``response.content`` and its
    metadata and the previously-generated fixture, if a diff is found then
    assertNoDiff() will raise a DiffsFound exception, printing out the diffs
    and commands used for the diffs, leaving temporary files behind for further
    investigation.
    """

    def __init__(self, path):
        """
        Instanciate a response object with a path to a fixture.

        Note that the ``for_test()`` class-method will generate a path.
        """
        self.path = path

    def assertNoDiff(self, response, selector=None):  # noqa
        """Backward compatibility method for pre-assertWebsiteSame versions."""
        diffs, created = self.make_diff(response, selector=selector)

        if created or diffs:
            raise DiffsFound(diffs, created)

    def make_diff(self, response, metadata=None, selector=None):
        """
        Compare a response object with the fixture.

        If the fixture doesn't exist, create it, otherwise run GNU-diff and
        return a list of diff outputs with their commands.

        Return created file list and dict of diffs.
        """
        if not os.path.exists(os.path.dirname(self.content_path)):
            os.makedirs(os.path.dirname(self.content_path))

        diffs = {}
        created = {}

        metadata = metadata or {}
        metadata['status_code'] = response.status_code

        if selector and not hasattr(response, 'streaming_content'):
            soup = BeautifulSoup(response.content, 'html5lib')
            elements = soup.select(selector)
            content = '\n---\n'.join(map(str, elements))
            mode = 'w+'
        else:
            try:
                content = response.content
            except AttributeError:
                content = b'\n'.join(
                    [i for i in response.streaming_content]
                )
            mode = 'wb+'

        if not os.path.exists(self.content_path):
            with open(self.content_path, mode) as f:
                f.write(content)
            created[self.content_path] = content

        if not os.path.exists(self.metadata_path):
            with open(self.metadata_path, 'w+') as f:
                json.dump(metadata, f, indent=4, sort_keys=True)
            created[self.metadata_path] = json.dumps(metadata, indent=4)

        fh, dump_path = tempfile.mkstemp('_responsediff')
        with os.fdopen(fh, mode) as f:
            f.write(content)

        cmd, out = diff(self.content_path, dump_path)
        if out:
            diffs[cmd] = out

        metadata_fh, metadata_dump_path = tempfile.mkstemp('_responsediff')
        with os.fdopen(metadata_fh, 'w') as f:
            json.dump(metadata, f, indent=4, sort_keys=True)

        cmd, out = diff(self.metadata_path, metadata_dump_path)
        if out:
            diffs[cmd] = out

        return diffs, created

    @property
    def metadata_path(self):
        """Return the path to the file for the response.metadata."""
        if self.path.endswith('/'):
            return os.path.join(self.path, 'metadata')
        return self.path + '.metadata'

    @property
    def content_path(self):
        """Return the path to the file for the response.content contents."""
        if self.path.endswith('/'):
            return os.path.join(self.path, 'content')
        return self.path + '.content'

    @classmethod
    def for_test(cls, case, url=None, *args, **kwargs):
        """Instanciate a Response with a path for the case and url if any."""
        name = '.'.join(case.id().split('.')[-2:])

        path = os.path.join(
            os.path.abspath(os.path.dirname(inspect.getfile(type(case)))),
            'response_fixtures',
            name
        )

        if url:
            path = os.path.join(
                path,
                *[p for p in url.split('/') if p]
            )

            if url.endswith('/'):
                path += '/'

        return cls(path, *args, **kwargs)



"""Convenience mixin for TestCases."""
import re

from django import test
from django.db import connections
from django.test.utils import CaptureQueriesContext

from .exceptions import DiffsFound
from .response import Response


class ResponseDiffTestMixin(object):
    """Adds assertResponseDiffEmpty() method."""

    def assertResponseDiffEmpty(self, result, selector=None):  # noqa
        """
        Test that result matches fixtures.

        When a selector is specified, the result will be parsed as HTML and
        only elements matching this selector will be tested.
        """
        Response.for_test(self).assertNoDiff(result, selector)

    def assertWebsiteSame(self, url=None, client=None, selector=None):  # noqa
        covered, diffs, created = self.responsediff_website_crawl(
            url, client, selector=selector)

        if created or diffs:
            raise DiffsFound(diffs, created)

        return covered

    def responsediff_website_crawl(self, url=None, client=None, covered=None,
                                   diffs=None, created=None, selector=None):
        """
        Test your website with one call to this method.

        It returns the list of covered URLs. But before, it tests that all
        fixtures for this test have been covered, or fails, requiring you the
        to remove obsolete files to succeed again.
        """
        url = url or '/'
        client = client or test.Client()
        covered = covered or []
        diffs = diffs or {}
        created = created or {}

        conn = connections['default']
        with CaptureQueriesContext(conn) as queries:
            response = client.get(url)
        self.process_response(response)
        metadata = {'query_count': len(queries)}

        _diffs, _created = Response.for_test(self, url).make_diff(
            response,
            metadata=metadata,
            # Don't apply selector on first url, so we do the layout once
            selector=selector if covered else None,
        )
        covered.append(url)
        created.update(_created)
        diffs.update(_diffs)

        if hasattr(response, 'streaming_content'):
            return covered, diffs, created

        results = re.findall(
            'href="((http://testserver)?/[^"]*)',
            response.content.decode('utf8')
        )

        for result in results:
            sub_url = re.sub(
                'http://testserver',
                '',
                result[0]
            )

            if sub_url in covered:
                continue

            if self.skip_url(sub_url):
                continue

            self.responsediff_website_crawl(
                sub_url,
                client=client,
                covered=covered,
                diffs=diffs,
                created=created,
                selector=selector,
            )

        return covered, diffs, created

    def get_content_replace_patterns(self, response):
        """Return a list of pattern:replacement for response contents."""
        return [
            ('\n.*csrfmiddlewaretoken.*\n', ''),
            ('\n.*webpack.bundle.*\n', ''),
        ]

    def process_response(self, response):
        """Taint the response before saving."""
        if hasattr(response, 'streaming_content'):
            return

        for replace in self.get_content_replace_patterns(response):
            response.content = re.sub(
                replace[0],
                replace[1],
                response.content.decode('utf8')
            ).encode('utf8')

    def skip_url(self, url):
        """Return true if the url should be skipped, skips STATIC_URL."""
        from django.conf import settings
        return url.startswith(settings.STATIC_URL)



import json
import os
import unittest

from django import test
from django.utils import six

from responsediff.exceptions import DiffsFound
from responsediff.response import Response


class TestResponseDiff(unittest.TestCase):
    def test_path(self):
        expected = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            'response_fixtures',
            'TestResponseDiff.test_path'
        )

        assert Response.for_test(self).path == expected

    def test_story(self):
        result = test.Client().get('/admin/')
        expected = Response.for_test(self)

        if os.path.exists(expected.content_path):  # pragma: no cover
            # Only makes sense when you're running the tests over and over
            # locally, makes no sense on single-usage containers like travis,
            # hence "no cover" above.
            os.unlink(expected.content_path)

        if os.path.exists(expected.metadata_path):  # pragma: no cover
            os.unlink(expected.metadata_path)

        with self.assertRaises(DiffsFound) as raises_result:
            expected.assertNoDiff(result)

        msg = (
            raises_result.exception.message
            if six.PY2 else raises_result.exception.args[0]
        )

        assert expected.content_path in msg
        assert expected.metadata_path in msg

        # should have been created
        assert os.path.exists(expected.content_path)

        # should pass now
        expected.assertNoDiff(result)

        with open(expected.content_path, 'w') as f:
            f.write('bla')

        with self.assertRaises(DiffsFound) as e:
            expected.assertNoDiff(result)

        expected_diff = '''
@@ -1 +1 @@
-bla
\ No newline at end of file
+<h1>Not Found</h1><p>The requested URL /admin/ was not found on this server.</p>
\ No newline at end of file
'''.lstrip()

        diff = e.exception.message if six.PY2 else e.exception.args[0]
        result_diff = '\n'.join(diff.split('\n')[2:])
        assert result_diff == expected_diff

        # Let's fix it and test status code now
        with open(expected.content_path, 'wb') as f:
            f.write(result.content)

        # Let's mess with the status code now
        with open(expected.metadata_path, 'w') as f:
            json.dump(
                {
                    'status_code': 418,  # RFC 2324 LOL
                },
                f,
                indent=4
            )

        with self.assertRaises(DiffsFound) as e:
            expected.assertNoDiff(result)



"""Exceptions for responsediff module."""


class ResponseDiffException(Exception):
    """Base exception for this app."""


class DiffsFound(ResponseDiffException):
    """Raised when a test has failed."""

    def __init__(self, diffs, created):
        """Exception for when a diff command had output."""
        message = [''] + [  # one empty line
            '[created] %s:\n%s' % (k, v[:60]) for k, v in created.items()
        ] + [
            '%s\n%s' % (cmd, out.decode('utf8'))
            for cmd, out in diffs.items()
        ]
        super(DiffsFound, self).__init__('\n'.join(message))



from setuptools import setup, find_packages
import os


# Utility function to read the README file.
# Used for the long_description. It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='django-responsediff',
    version='0.7.2',
    description='HTTP response diffing against fixtures for testing',
    author='James Pic',
    author_email='jamespic@gmail.com',
    url='https://github.com/yourlabs/django-responsediff',
    packages=find_packages(),
    include_package_data=True,
    long_description=read('README.rst'),
    license='MIT',
    keywords='django test response fixture diff',
    install_requires=[
        'beautifulsoup4',
        'html5lib',
        'six'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
